package edu.rice.cs.mm;

import java.lang.reflect.*;
import java.util.*;

public class MMProxy implements java.lang.reflect.InvocationHandler {
    private Object _obj;
    private Class<?> _cl;
    private boolean _clMultiMethodOnClass = false;
    private boolean _clHasCheckedInterfaces = false;
    private HashMap<Method,Boolean> _clMultiMethodOnMethod = new HashMap<Method,Boolean>();
    
    public static final boolean THROW_IF_NOT_UNIQUE = false;
    public static final boolean DEBUG = false;
    
    private static void debugOut(String s) {
        if (DEBUG) { System.err.println(s); }
    }
    
    @SuppressWarnings("unchecked")
    public static <T> T newInstance(T obj) {
        return (T)java.lang.reflect.Proxy.newProxyInstance(obj.getClass().getClassLoader(),
                                                           obj.getClass().getInterfaces(),
                                                           new MMProxy(obj));
    }
    
    private MMProxy(Object obj) {
        _obj = obj;
        _cl = obj.getClass();
        _clMultiMethodOnClass = (_cl.getAnnotation(MultiMethod.class)!=null);
    }
    
    private static int getTypeDistance(Class c0, Class c) {
//        debugOut("getTypeDistance("+c0+", "+c+")");
        if (c0==null) { return 0; }
        if (c0.equals(c)) { return 0; }
        int min = getTypeDistance(c0.getSuperclass(), c);
        for(Class i: c0.getInterfaces()) {
            int d = getTypeDistance(i, c);
            if (d<min) { min = d; }
        }
        return min+1;
    }
    
    @SuppressWarnings("unchecked")
    private static boolean checkForAnnotationInInterfaces(Class c) {
        for(Class i: c.getInterfaces()) {
            if (i.getAnnotation(MultiMethod.class)!=null) return true;
            if (checkForAnnotationInInterfaces(i)) return true;
        }
        if (c.getSuperclass()!=null) return checkForAnnotationInInterfaces(c.getSuperclass());
        return false;
    }
    
    private static boolean checkForAnnotationOnMethods(Class c, Method m) {
        // check method in this class
        try {
            Method mThis = c.getMethod(m.getName(), m.getParameterTypes());
            if (mThis.getAnnotation(MultiMethod.class)!=null) return true;
        }
        catch(NoSuchMethodException e) { /* ignore */ }
        Class cSup = c.getSuperclass();
        if (cSup!=null) {
            // check method in superclass
            try {
                Method mSup = cSup.getMethod(m.getName(), m.getParameterTypes());
                if (checkForAnnotationOnMethods(cSup, mSup)) return true;
            }
            catch(NoSuchMethodException e) { /* do not recurse deeper */ }
        }
        // check interfaces
        for(Class i: c.getInterfaces()) {
            // check method in superclass
            try {
                Method mSup = i.getMethod(m.getName(), m.getParameterTypes());
                if (checkForAnnotationOnMethods(i, mSup)) return true;
            }
            catch(NoSuchMethodException e) { /* do not recurse deeper */ }
        }        
        return false;
    }
    
    public Object invoke(Object proxy, Method m, Object[] args) throws Throwable {
        debugOut("MMProxy.invoke: m = "+m+", args = "+java.util.Arrays.toString(args));
        Object result;
        final String mName = m.getName();
        final int mMod = m.getModifiers();
        final Class[] mParam = m.getParameterTypes();
        if (args.length!=mParam.length) {
            throw new IllegalArgumentException("Incorrect number of arguments passed: method "+m+" requires "+mParam.length+" arguments, "+
                                               "but only "+args.length+" arguments were passed, "+java.util.Arrays.toString(args));
        }
        Method mm = m;
        mm.setAccessible(true);
        boolean doMulti = _clMultiMethodOnClass;
        if (!doMulti) {
            if (!_clHasCheckedInterfaces) {
                // check for annotation in interfaces
                doMulti = _clMultiMethodOnClass = checkForAnnotationInInterfaces(_cl);
                _clHasCheckedInterfaces = true;
            }
        }
        if (!doMulti) {
            Boolean b = _clMultiMethodOnMethod.get(m);
            if (b==null) {
                // check for annotation on method in superclasses and interfaces
                b = checkForAnnotationOnMethods(_cl, m);
                _clMultiMethodOnMethod.put(m, b);
            }
            doMulti = b;
        }
        if (doMulti) {
            // do multiple dispatch
            debugOut("Multiple dispatch");
            Method[] msarr = _cl.getDeclaredMethods();
            LinkedList<Method> ms = new LinkedList<Method>();
            for(Method marr: msarr) {
                if ((marr.getName().equals(mName)) &&                 // same name and
                    (marr.getParameterTypes().length==mParam.length)) { // same number of parameters
                    if ((mMod & Modifier.PUBLIC)!=0) { // method is public
                        // allow public methods
                        if ((marr.getModifiers() & Modifier.PUBLIC)!=0) { ms.add(marr); }
                    }
                    else if ((mMod & Modifier.PROTECTED)!=0) { // method is protected
                        // allow public and protected methods
                        if ((marr.getModifiers() & (Modifier.PUBLIC|Modifier.PROTECTED))!=0) { ms.add(marr); }
                    }
                    else if ((mMod & Modifier.PRIVATE)!=0) { // method is private
                        // allow public, protected and private methods
                        if ((marr.getModifiers() & (Modifier.PUBLIC|Modifier.PROTECTED|Modifier.PRIVATE))!=0) { ms.add(marr); }
                    }
                    else { // method is package protected
                        // allow public, protected and package protected methods
                        if (((marr.getModifiers() & (Modifier.PUBLIC|Modifier.PROTECTED))!=0) &&
                            ((marr.getModifiers() & (Modifier.PRIVATE))==0)) { ms.add(marr); }
                    }
                }
            }
            TreeMap<Integer,ArrayList<Method>> msFiltered = new TreeMap<Integer,ArrayList<Method>>();
            for(Method mcand: ms) {
                debugOut("\t"+mcand);
                boolean match = true;
                int i = 0;
                int dist = 0;
                for(Class c: mcand.getParameterTypes()) {
                    if (!c.isInstance(args[i])) {
                        match = false;
                        break;
                    }
                    dist += getTypeDistance(args[i].getClass(), c);
                    ++i;
                }
                if (match) {
                    ArrayList<Method> mList;
                    if (msFiltered.containsKey(dist)) {
                        mList = msFiltered.get(dist);
                    }
                    else {
                        mList = new ArrayList<Method>();
                    }
                    mList.add(mcand);
                    msFiltered.put(dist,mList);
                }
            }
            debugOut("\t----ArgTypes----");
            for(Object arg: args) { debugOut("\t"+arg.getClass()); }
            debugOut("\t----Filtered----");
            for(Map.Entry<Integer,ArrayList<Method>> mcand: msFiltered.entrySet()) {
                debugOut("\t"+mcand.getKey()+": "+mcand.getValue());
            }
            ArrayList<Method> bestMatches = msFiltered.get(msFiltered.firstKey());
            debugOut("\t----Best Matches----");
            debugOut("\t"+bestMatches);
            if (bestMatches.size()!=1) {
                debugOut("\t====NOT A UNIQUE MATCH====");
                if (THROW_IF_NOT_UNIQUE) {
                    throw new NoSuchMethodException("There was no unique method that matched the the types of the parameters "+
                                                    java.util.Arrays.toString(args)+"; equally good candidate methods were: "+
                                                    bestMatches);
                }
            }
            mm = bestMatches.get(0);
        }
        else { debugOut("Single dispatch"); }
        try {
            debugOut("\t===CALLING METHOD: "+mm);
            mm.setAccessible(true);
            result = mm.invoke(_obj, args);
        }
        catch (InvocationTargetException e) {
            throw e.getTargetException();
        }
        return result;
    }
}